"""
    lager.erase.commands

    Commands for erasing a DUT
"""
import click
from ..context import get_default_gateway
from ..util import stream_output

def do_reset(session, gateway, halt):
    """
        Perform the DUT reset
    """
    return session.reset_dut(gateway, halt)

@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.option('--halt', is_flag=True, help='Halt the DUT after reset. Default: do not halt', default=False, show_default=True)
def reset(ctx, gateway, halt):
    """
        Reset the DUT. By default will not halt the DUT; use `lager reset --halt` to reset and halt.
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)
    resp = do_reset(ctx.obj.session, gateway, halt)
    stream_output(resp)



"""
    lager.uart.commands

    Commands for DUT UART interaction
"""
import sys
import math
import click
from ..context import get_default_gateway
from ..status import run_job_output
from ..config import read_config_file

def do_uart(ctx, gateway, serial_device, baudrate, bytesize, parity, stopbits, xonxoff, rtscts, dsrdtr, test_runner):
    if serial_device is None:
        config = read_config_file()
        if 'LAGER' in config and 'serial_device' in config['LAGER']:
            serial_device = config.get('LAGER', 'serial_device')

    if not serial_device:
        raise click.UsageError(
            '--serial-device required',
            ctx=ctx,
        )

    if xonxoff and rtscts:
        raise click.UsageError(
            'Cannot use xonxoff and rtscts simultaneously',
            ctx=ctx,
        )
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session
    serial_options = {
        'device': serial_device,
        'baudrate': baudrate,
        'bytesize': bytesize,
        'parity': parity,
        'stopbits': stopbits,
        'xonxoff': xonxoff,
        'rtscts': rtscts,
        'dsrdtr': dsrdtr,
    }
    return session.uart_gateway(gateway, serial_options, test_runner)


@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.option('--serial-device', help='Gateway serial port device path', metavar='path')
@click.option('--baudrate', help='Serial baud rate', type=int, default=None)
@click.option('--bytesize', help='Number of data bits', type=click.Choice(['5', '6', '7', '8']), default=None)
@click.option('--parity', help='Parity check', type=click.Choice(['none', 'even', 'odd', 'mark', 'space']), default=None)
@click.option('--stopbits', help='Number of stop bits', type=click.Choice(['1', '1.5', '2']), default=None)
@click.option('--xonxoff/--no-xonxoff', default=None, help='Enable/disable software XON/XOFF flow control')
@click.option('--rtscts/--no-rtscts', default=None, help='Enable/disable hardware RTS/CTS flow control')
@click.option('--dsrdtr/--no-dsrdtr', default=None, help='Enable/disable hardware DSR/DTR flow control')
@click.option('--test-runner', help='End the UART session when end-of-test is detected', type=click.Choice(['none', 'unity']), default='unity')
@click.option('--interactive', is_flag=True, help='Run as an interactive TTY session', default=False)
@click.option('--message-timeout', default=math.inf, type=click.FLOAT,
              help='Max time in seconds to wait between messages from API.')
@click.option('--overall-timeout', default=math.inf, type=click.FLOAT,
              help='Cumulative time in seconds to wait for session output.')
@click.option('--eof-timeout', default=None, type=click.FLOAT,
              help='Time in seconds to wait before closing connection after input EOF received')
@click.option('--display-job-id', default=False, is_flag=True)
def uart(ctx, gateway, serial_device, baudrate, bytesize, parity, stopbits, xonxoff, rtscts, dsrdtr,
         test_runner, interactive, message_timeout, overall_timeout, eof_timeout, display_job_id):
    """
        Connect to UART on a DUT.
    """
    if interactive:
        if not sys.stdin.isatty():
            click.echo('stdin is not a tty!', err=True)
            ctx.exit(1)
        if not sys.stdout.isatty():
            click.echo('stdout is not a tty!', err=True)
            ctx.exit(1)

    resp = do_uart(
        ctx, gateway, serial_device, baudrate, bytesize, parity,
        stopbits, xonxoff, rtscts, dsrdtr, test_runner,
    )
    test_run = resp.json()
    job_id = test_run['test_run']['id']
    if display_job_id:
        click.echo('Job id: {}'.format(job_id), err=True)

    connection_params = ctx.obj.websocket_connection_params(socktype='job', job_id=job_id)
    run_job_output(connection_params, test_runner, interactive, message_timeout, overall_timeout,
        eof_timeout, ctx.obj.debug)



"""
    lager.flash.commands

    Commands for flashing a DUT
"""
import itertools
import click
from ..context import get_default_gateway
from ..util import stream_output
from ..paramtypes import BinfileType

def do_flash(session, gateway, hexfile, binfile, preverify, verify, force=False):
    """
        Perform the actual flash operation
    """
    files = list(zip(itertools.repeat('hexfile'), [open(path, 'rb') for path in hexfile]))
    files.extend(
        zip(itertools.repeat('binfile'), [open(binf.path, 'rb') for binf in binfile])
    )
    files.extend(
        zip(itertools.repeat('binfile_address'), [binf.address for binf in binfile])
    )
    files.append(('preverify', preverify))
    files.append(('verify', verify))
    files.append(('force', force))

    return session.flash_dut(gateway, files=files)


@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.option(
    '--hexfile',
    multiple=True, type=click.Path(exists=True),
    help='Hexfile(s) to flash. May be passed multiple times; files will be flashed in order.')
@click.option(
    '--binfile',
    multiple=True, type=BinfileType(exists=True),
    help='Binfile(s) to flash. Syntax: --binfile `<filename>,<address>` '
         'May be passed multiple times; files will be flashed in order.')
@click.option(
    '--preverify/--no-preverify',
    help='If true, only flash target if image differs from current flash contents',
    default=True, show_default=True)
@click.option('--verify/--no-verify', help='Verify image successfully flashed', default=True, show_default=True)
@click.option('--force', is_flag=True)
def flash(ctx, gateway, hexfile, binfile, preverify, verify, force):
    """
        Flash a DUT connected to a gateway with 1 or more bin or hex files
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session

    resp = do_flash(session, gateway, hexfile, binfile, preverify, verify, force)
    stream_output(resp)



"""
    lager.gateway.commands

    Gateway commands
"""
import click
from ..config import read_config_file, write_config_file

@click.group(name='set')
def setter():
    """
        Lager setter commands
    """
    pass

@setter.group()
def default():
    """
        Set defaults
    """
    pass

@default.command()
@click.argument('gateway_id')
def gateway(gateway_id):
    """
        Set default gateway
    """
    config = read_config_file()
    config.set('LAGER', 'gateway_id', gateway_id)
    write_config_file(config)

@default.command()
@click.argument('device_path')
def serial_device(device_path):
    """
        Set default serial device path
    """
    config = read_config_file()
    config.set('LAGER', 'serial_device', device_path)
    write_config_file(config)



"""
    lager.gdbserver.commands

    GDB Server tunnel commands
"""

import click
import trio
from .tunnel import serve_tunnel, serve_local_tunnel
from ..context import get_default_gateway, ensure_debugger_running

def _run_gdbserver_cloud(ctx, host, port, gateway):
    connection_params = ctx.obj.websocket_connection_params(socktype='gdb-tunnel', gateway_id=gateway)
    try:
        trio.run(serve_tunnel, host, port, connection_params)
    except PermissionError as exc:
        if port < 1024:
            click.secho(f'Permission denied for port {port}. Using a port number less than '
                        '1024 typically requires root privileges.', fg='red', err=True)
        else:
            click.secho(str(exc), fg='red', err=True)
        if ctx.obj.debug:
            raise
    except OSError as exc:
        click.secho(f'Could not start gdbserver on port {port}: {exc}', fg='red', err=True)
        if ctx.obj.debug:
            raise

def _run_gdbserver_local(ctx, host, port, gateway, fork):
    try:
        trio.run(serve_local_tunnel, ctx.obj.session, gateway, host, port, fork)
    except PermissionError as exc:
        if port < 1024:
            click.secho(f'Permission denied for port {port}. Using a port number less than '
                        '1024 typically requires root privileges.', fg='red', err=True)
        else:
            click.secho(str(exc), fg='red', err=True)
        if ctx.obj.debug:
            raise
    except OSError as exc:
        click.secho(f'Could not start gdbserver on port {port}: {exc}', fg='red', err=True)
        if ctx.obj.debug:
            raise

@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.option('--host', default='localhost', help='interface for gdbserver to bind. '
              'Use --host \'*\' to bind to all interfaces.', show_default=True)
@click.option('--port', default=3333, help='Port for gdbserver', show_default=True)
@click.option('--local', is_flag=True, default=False, help='Connect to gateway via local network', show_default=True)
@click.option('--fork', is_flag=True, default=False, help='Allow forking', show_default=True)
def gdbserver(ctx, gateway, host, port, local, fork):
    """
        Establish a proxy to GDB server on gateway. By default binds to localhost, meaning gdb
        client connections must originate from the machine running `lager gdbserver`. If you would
        like to bind to all interfaces, use --host '*'

        The --local flag can be used if you are on the same network as your gateway, and will cause
        `lager gdbserver` to directly connect to your gateway on the local network for reduced latency.
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    ensure_debugger_running(gateway, ctx)

    if local:
        _run_gdbserver_local(ctx, host, port, gateway, fork)
    else:
        _run_gdbserver_cloud(ctx, host, port, gateway)



"""
    lager.paramtypes

    Custom click paramtypes
"""
import collections
import os
import click

class MemoryAddressType(click.ParamType):
    """
        Memory address integer parameter
    """
    name = 'memory address'

    def convert(self, value, param, ctx):
        """
            Parse string reprsentation of a hex integer
        """
        value = value.strip().lower()
        if value.lower().startswith('0x'):
            try:
                return int(value, 16)
            except ValueError:
                self.fail(f"{value} is not a valid hex integer", param, ctx)

        try:
            return int(value, 10)
        except ValueError:
            self.fail(f"{value} is not a valid integer", param, ctx)

    def __repr__(self):
        return 'ADDR'

class HexParamType(click.ParamType):
    """
        Hexadecimal integer parameter
    """
    name = 'hex'

    def convert(self, value, param, ctx):
        """
            Parse string reprsentation of a hex integer
        """
        try:
            return int(value, 16)
        except ValueError:
            self.fail(f"{value} is not a valid hex integer", param, ctx)

    def __repr__(self):
        return 'HEX'

class VarAssignmentType(click.ParamType):
    """
        Hexadecimal integer parameter
    """
    name = 'FOO=BAR'

    def convert(self, value, param, ctx):
        """
            Parse string reprsentation of a hex integer
        """
        parts = value.split('=')
        if len(parts) != 2:
            raise ValueError('Invalid assignment')

        return parts

    def __repr__(self):
        return 'VAR ASSIGNMENT'

Binfile = collections.namedtuple('Binfile', ['path', 'address'])
class BinfileType(click.ParamType):
    """
        Type to represent a command line argument for a binfile (<path>,<address>)
    """
    envvar_list_splitter = os.path.pathsep
    name = 'binfile'

    def __init__(self, *args, exists=False, **kwargs):
        self.exists = exists
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        """
            Convert binfile param string into useable components
        """
        parts = value.rsplit(',', 1)
        if len(parts) != 2:
            self.fail(f'{value}. Syntax: --binfile <filename>,<address>', param, ctx)
        filename, address = parts
        path = click.Path(exists=self.exists).convert(filename, param, ctx)
        address = HexParamType().convert(address, param, ctx)

        return Binfile(path=path, address=address)

    def __repr__(self):
        return 'BINFILE'



"""
    lager_cli.util

    Catchall for utility functions
"""
import sys
import click
import trio
import lager_trio_websocket as trio_websocket
import wsproto.frame_protocol as wsframeproto

def stream_output(response, chunk_size=1):
    """
        Stream an http response to stdout
    """
    for chunk in response.iter_content(chunk_size=chunk_size):
        click.echo(chunk, nl=False)
        sys.stdout.flush()

async def heartbeat(websocket, timeout, interval):
    '''
    Send periodic pings on WebSocket ``ws``.

    Wait up to ``timeout`` seconds to send a ping and receive a pong. Raises
    ``TooSlowError`` if the timeout is exceeded. If a pong is received, then
    wait ``interval`` seconds before sending the next ping.

    This function runs until cancelled.

    :param ws: A WebSocket to send heartbeat pings on.
    :param float timeout: Timeout in seconds.
    :param float interval: Interval between receiving pong and sending next
        ping, in seconds.
    :raises: ``ConnectionClosed`` if ``ws`` is closed.
    :raises: ``TooSlowError`` if the timeout expires.
    :returns: This function runs until cancelled.
    '''
    try:
        while True:
            with trio.fail_after(timeout):
                await websocket.ping()
            await trio.sleep(interval)
    except trio_websocket.ConnectionClosed as exc:
        if exc.reason is None:
            return
        if exc.reason.code != wsframeproto.CloseReason.NORMAL_CLOSURE or exc.reason.reason != 'EOF':
            raise



"""
    lager.gateway.commands

    Gateway commands
"""
import collections
import os
import click
from ..context import get_default_gateway

@click.group(name='gateway')
def _gateway():
    """
        Lager gateway commands
    """
    pass

@_gateway.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
def hello(ctx, gateway):
    """
        Say hello to gateway
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)
    session = ctx.obj.session
    resp = session.gateway_hello(gateway)
    click.echo(resp.content, nl=False)

@_gateway.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.option('--model', required=False)
def serial_numbers(ctx, gateway, model):
    """
        Get serial numbers of devices attached to gateway
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session
    resp = session.serial_numbers(gateway, model)
    for device in resp.json()['devices']:
        device['serial'] = device['serial'].lstrip('0')
        click.echo('{vendor} {model}: {serial}'.format(**device))

@_gateway.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
def serial_ports(ctx, gateway):
    """
        Get serial ports attached to gateway
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session
    resp = session.serial_ports(gateway)
    style = ctx.obj.style
    for port in resp.json()['serial_ports']:
        click.echo('{} - {}'.format(style(port['device'], fg='green'), port['description']))

class HexParamType(click.ParamType):
    """
        Hexadecimal integer parameter
    """
    name = 'hex'

    def convert(self, value, param, ctx):
        """
            Parse string reprsentation of a hex integer
        """
        try:
            return int(value, 16)
        except ValueError:
            self.fail(f"{value} is not a valid hex integer", param, ctx)

    def __repr__(self):
        return 'HEX'

Binfile = collections.namedtuple('Binfile', ['path', 'address'])
class BinfileType(click.ParamType):
    """
        Type to represent a command line argument for a binfile (<path>,<address>)
    """
    envvar_list_splitter = os.path.pathsep
    name = 'binfile'

    def __init__(self, *args, exists=False, **kwargs):
        self.exists = exists
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        """
            Convert binfile param string into useable components
        """
        parts = value.rsplit(',', 1)
        if len(parts) != 2:
            self.fail(f'{value}. Syntax: --binfile <filename>,<address>', param, ctx)
        filename, address = parts
        path = click.Path(exists=self.exists).convert(filename, param, ctx)
        address = HexParamType().convert(address, param, ctx)

        return Binfile(path=path, address=address)

    def __repr__(self):
        return 'BINFILE'

@_gateway.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
def status(ctx, gateway):
    """
        Get gateway debugger status
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session
    response = session.gateway_status(gateway).json()
    running = response['running']
    cmdline = response['cmdline']
    logfile = response['logfile']
    click.echo(f'Debugger running: {running}')
    if cmdline:
        click.echo('---- Debugger config ----')
        config = cmdline[3:-1]
        for i in range(0, len(config), 2):
            chunk = config[i:i+2]
            if chunk[0] == '-f':
                parts = chunk[1].split('/')
                click.echo(f'{parts[0]}: {parts[-1].rstrip(".cfg")}')
            elif chunk[0] == '-c':
                click.echo(chunk[1])

    if logfile:
        click.echo('---- Logs ----')
        click.echo(logfile)


@_gateway.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to rename')
@click.option('--to', required=True, help='New name for gateway')
def rename(ctx, gateway, to):
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session
    session.rename_gateway(gateway, to)



"""
    lager.auth

    Authentication helpers
"""
import os
import base64
import json
import time
import datetime
import urllib.parse
import requests
from ..config import read_config_file, write_config_file

_DEFAULT_CLIENT_ID = 'N2NcC4UplFLlxiIPzDDa5K1PKbTrRVec'
_DEFAULT_AUDIENCE = 'https://app.lagerdata.com/gateway'
_DEFAULT_AUTH_URL = 'https://auth.lagerdata.com'

_JWK_PATH = '.well-known/jwks.json'

_EXPIRY_GRACE = datetime.timedelta(hours=1).seconds

def get_auth_url():
    return os.getenv('LAGER_AUTH_URL', _DEFAULT_AUTH_URL)

def get_client_id():
    return os.getenv('LAGER_CLIENT_ID', _DEFAULT_CLIENT_ID)

def get_audience():
    return os.getenv('LAGER_AUDIENCE', _DEFAULT_AUDIENCE)

def _get_jwks(jwk_url):
    resp = requests.get(jwk_url)
    resp.raise_for_status()
    return resp.json()

def _is_expired(token):
    try:
        _header, encoded_payload, _sig = token.split('.')
        padding = '===='  # Auth0 token does not include padding
        payload = json.loads(base64.b64decode(encoded_payload + padding))
        return payload['exp'] < time.time() + _EXPIRY_GRACE
    except ValueError:
        return True

def _refresh(refresh_token):
    data = {
        'grant_type': 'refresh_token',
        'client_id': get_client_id(),
        'refresh_token': refresh_token,
    }
    token_url = urllib.parse.urljoin(get_auth_url(), '/oauth/token')
    resp = requests.post(token_url, data=data)
    resp.raise_for_status()
    return resp.json()

AUTH_TOKEN_KEY = 'LAGER_AUTH_TOKEN'
REFRESH_TOKEN_KEY = 'LAGER_REFRESH_TOKEN'
TOKEN_TYPE_KEY = 'LAGER_TOKEN_TYPE'
SECRET_TOKEN_KEY = 'LAGER_SECRET_TOKEN'

def _load_auth_from_environ(env):
    if SECRET_TOKEN_KEY in env:
        return dict(
            token=env[SECRET_TOKEN_KEY],
            type='secret',
        )

    if AUTH_TOKEN_KEY in env and TOKEN_TYPE_KEY in env and (
            env[TOKEN_TYPE_KEY].lower() == 'secret' or REFRESH_TOKEN_KEY in env):
        return dict(
            token=env[AUTH_TOKEN_KEY],
            refresh=env.get(REFRESH_TOKEN_KEY),
            type=env[TOKEN_TYPE_KEY],
        )
    return None

def load_auth():
    """
        Load auth token from environment if available, otherwise config
    """
    auth = None
    update_config = False
    auth = _load_auth_from_environ(os.environ)
    if not auth:
        config = read_config_file()
        if 'AUTH' not in config or 'token' not in config['AUTH']:
            return None
        update_config = True
        auth = config['AUTH']

    if auth['type'].lower() == 'bearer' and _is_expired(auth['token']):
        fresh_token = _refresh(auth['refresh'])
        auth['token'] = fresh_token['access_token']
        auth['type'] = fresh_token['token_type']
        if update_config:
            write_config_file(config)

    return auth



"""
    lager.status

    Job status output functions
"""
try:
    _CURSES_IMPORT_FAILED = False
    import curses
except (ImportError, ModuleNotFoundError) as exc:
    _CURSES_IMPORT_FAILED = exc
import platform
import threading
import sys
import select
from functools import partial
import bson
import click
import trio
import lager_trio_websocket as trio_websocket
from lager_trio_websocket import open_websocket_url
from tenacity import retry, stop_after_attempt, wait_fixed, retry_if_exception_type
import wsproto.frame_protocol as wsframeproto
import requests
from .matchers import test_matcher_factory
from .util import heartbeat

def stream_response(response):
    """
        Stream a ``requests`` response to the terminal
    """
    with response:
        for chunk in response.iter_content(chunk_size=None):
            click.echo(chunk, nl=False)

async def handle_url_message(matcher, urls):
    """
        Handle a message with data location urls
    """
    downloader = partial(requests.get, stream=True)
    for url in urls:
        response = await trio.to_thread.run_sync(downloader, url)
        response.raise_for_status()
        await trio.to_thread.run_sync(stream_response, response)

async def handle_data_message(matcher, message):
    """
        Handle a data message
    """
    for item in message:
        entry = item['entry']
        if 'payload' in entry:
            payload = entry['payload']
            matcher.feed(payload)

async def handle_message(matcher, message):
    """
        Handle an individual parsed websocket message
    """
    if 'data' in message:
        return await handle_data_message(matcher, message['data'])
    if 'urls' in message:
        return await handle_url_message(matcher, message['urls'])
    return None

class InterMessageTimeout(Exception):
    """
        Raised if no messages received for ``message_timeout`` seconds
    """
    pass

async def read_from_websocket(websocket, matcher, message_timeout, nursery):
    try:
        while True:
            try:
                with trio.fail_after(message_timeout):
                    try:
                        message = await websocket.get_message()
                    except trio_websocket.ConnectionClosed as exc:
                        if exc.reason is None:
                            return
                        if exc.reason.code != wsframeproto.CloseReason.NORMAL_CLOSURE or exc.reason.reason != 'EOF':
                            raise
                        break
            except trio.TooSlowError:
                raise InterMessageTimeout(message_timeout)
            await handle_message(matcher, bson.loads(message))
    finally:
        matcher.done()
        nursery.cancel_scope.cancel()

def reader_function(io, send_channel, trio_token):
    if platform.system() == 'Windows':
        return
    try:
        while True:
            data = io.read()
            if data is not None:
                trio.from_thread.run(send_channel.send, {'type': 'data', 'value': data}, trio_token=trio_token)
    except EOFError:
        trio.from_thread.run(send_channel.send, {'type': 'EOF'}, trio_token=trio_token)


async def write_to_websocket(websocket, receive_channel, eof_timeout, nursery):
    while True:
        message = await receive_channel.receive()
        if message['type'] == 'EOF':
            if eof_timeout is not None:
                await trio.sleep(eof_timeout)
                await websocket.aclose()
            return

        if message['type'] == 'data':
            data = message['value']
            try:
                await websocket.send_message(data)
            except trio_websocket.ConnectionClosed as exc:
                if exc.reason is None:
                    return
                if exc.reason.code != wsframeproto.CloseReason.NORMAL_CLOSURE or exc.reason.reason != 'EOF':
                    raise
                return

class TTYIO:
    """
        Class for doing I/O with a TTY managed by curses
    """
    def __init__(self):
        stdscr = curses.initscr()
        curses.noecho()
        curses.cbreak()
        stdscr.keypad(True)
        self.stdscr = stdscr

    def shutdown(self):
        """
            Restore previous TTY settings
        """
        curses.nocbreak()
        self.stdscr.keypad(False)
        curses.echo()
        curses.endwin()

    def output(self, data):
        """
            Output some data to the TTY
        """
        self.stdscr.addstr(data)
        self.stdscr.refresh()

    def read(self):
        """
            Read a key from the TTY
        """
        char = self.stdscr.getch()
        if char == -1:
            raise EOFError
        if char <= 255:
            return char.to_bytes(1, byteorder='little')
        return None

class StandardIO:
    """
        Class for doing I/O with standard UNIX io streams
    """
    def shutdown(self):
        """
            No special behavior needed
        """
        pass

    def output(self, data):
        """
            Send some data to stdout
        """
        click.echo(data, nl=False)

    def read(self):
        """
            Read some data from stdin
        """
        rlist, _wlist, xlist = select.select([sys.stdin], [], [sys.stdin])
        if rlist:
            chunk = sys.stdin.buffer.read()
            if chunk == b'':
                raise EOFError
            return chunk
        if xlist:
            raise EOFError
        return None


@retry(reraise=True, sleep=trio.sleep, stop=stop_after_attempt(4), wait=wait_fixed(2), retry=retry_if_exception_type(trio_websocket.ConnectionRejected))
async def display_job_output(connection_params, test_runner, interactive, message_timeout, overall_timeout, eof_timeout):
    """
        Display job output from websocket
    """
    (uri, kwargs) = connection_params
    match_class = test_matcher_factory(test_runner)
    if interactive:
        io_source = TTYIO()
    else:
        io_source = StandardIO()
    try:
        matcher = match_class(io_source)
        send_channel, receive_channel = trio.open_memory_channel(0)
        with trio.fail_after(overall_timeout):
            async with open_websocket_url(uri, disconnect_timeout=1, **kwargs) as websocket:
                token = trio.lowlevel.current_trio_token()
                thread = threading.Thread(target=reader_function, args=(io_source, send_channel, token), daemon=True)
                thread.start()
                async with trio.open_nursery() as nursery:
                    nursery.start_soon(heartbeat, websocket, 30, 30)
                    nursery.start_soon(read_from_websocket, websocket, matcher, message_timeout, nursery)
                    nursery.start_soon(write_to_websocket, websocket, receive_channel, eof_timeout, nursery)
        return matcher
    finally:
        if io_source:
            io_source.shutdown()

def run_job_output(connection_params, test_runner, interactive, message_timeout, overall_timeout, eof_timeout, debug=False):
    """
        Run async task to get job output from websocket
    """
    if interactive and _CURSES_IMPORT_FAILED:
        click.echo(_CURSES_IMPORT_FAILED, err=True)
        click.echo('Please install curses module before using --interactive')
        click.get_current_context().exit(1)

    try:
        matcher = trio.run(display_job_output, connection_params, test_runner, interactive, message_timeout, overall_timeout, eof_timeout)
        click.get_current_context().exit(matcher.exit_code)
    except trio.TooSlowError:
        suffix = '' if overall_timeout == 1 else 's'
        message = f'Job status timed out after {overall_timeout} second{suffix}'
        click.secho(message, fg='red', err=True)
        click.get_current_context().exit(1)
    except InterMessageTimeout:
        suffix = '' if message_timeout == 1 else 's'
        message = f'Timed out after no messages received for {message_timeout} second{suffix}'
        click.secho(message, fg='red', err=True)
        click.get_current_context().exit(1)
    except requests.exceptions.HTTPError as exc:
        response = getattr(exc, 'response')
        if response is not None and response.status_code == 404:
            click.secho('Test run content not found', fg='red', err=True)
        else:
            click.secho('Error retrieving test run content', fg='red', err=True)
        if debug:
            raise
    except trio_websocket.ConnectionRejected as exc:
        if exc.status_code == 404:
            click.secho('Job not found', fg='red', err=True)
        elif exc.status_code >= 500:
            click.secho('Internal error in Lager API. '
                        'Please contact support@lagerdata.com if this persists.',
                        fg='red', err=True)
        else:
            click.secho('Could not connect to API websocket', fg='red', err=True)
        if debug:
            raise
        click.get_current_context().exit(1)
    except trio_websocket.HandshakeError as exc:
        click.secho('Could not connect to API websocket', fg='red', err=True)
        if debug:
            raise
        click.get_current_context().exit(1)
    except trio_websocket.ConnectionClosed as exc:
        if exc.reason.code != wsframeproto.CloseReason.NORMAL_CLOSURE:
            click.secho('API websocket closed abnormally', fg='red', err=True)
            if debug:
                raise
            click.get_current_context().exit(1)
        elif exc.reason.reason != 'EOF':
            click.secho('API websocket closed unexpectedly', fg='red', err=True)
            if debug:
                raise
            click.get_current_context().exit(1)
    except ConnectionRefusedError:
        click.secho('Lager API websocket connection refused!', fg='red', err=True)
        if debug:
            raise
        click.get_current_context().exit(1)



"""
    lager.exec.commands

    Run commands in a docker container
"""
import subprocess
import os
import click
from ..config import (
    write_config_file,
    add_devenv_command,
    get_devenv_config,
    DEVENV_SECTION_NAME,
)
from ..context import get_ci_environment, CIEnvironment

def _run_command_host(section, path, cmd_to_run, extra_args, debug):
    """
        Run a command from the host (which means, run it in a docker container)
    """
    full_command = ' '.join((cmd_to_run, *extra_args)).strip()
    if full_command.startswith('lager'):
        proc = subprocess.run(full_command, shell=True, check=False)
        return proc.returncode

    image = section.get('image')
    source_dir = os.path.dirname(path)
    mount_dir = section.get('mount_dir')
    shell = section.get('shell')
    if debug:
        click.echo(full_command, err=True)
    env_vars = [var for var in os.environ if var.startswith('LAGER')]
    env_strings = [f'--env={var}={os.environ[var]}' for var in env_vars]
    base_command = ['docker', 'run', '-it', '--rm']
    base_command.extend(env_strings)
    base_command.extend([
        '-v',
        f'{source_dir}:{mount_dir}',
        image,
        shell,
        '-c',
        full_command
    ])
    proc = subprocess.run(base_command, check=False)
    return proc.returncode

def _run_command_container(section, cmd_to_run, extra_args, debug):
    """
        Run a command directly - assume we are in a container with all necessary software
        installed already.
    """
    shell = section.get('shell')
    full_command = ' '.join((cmd_to_run, *extra_args))
    if debug:
        click.echo(full_command, err=True)
    proc = subprocess.run([shell, '-c', full_command], check=False)
    return proc.returncode

def _run_command(section, path, cmd_to_run, extra_args, debug):
    ci_env = get_ci_environment()
    if ci_env == CIEnvironment.CONTAINER_CI:
        return _run_command_container(section, cmd_to_run, extra_args, debug)
    if ci_env == CIEnvironment.HOST:
        return _run_command_host(section, path, cmd_to_run, extra_args, debug)
    raise ValueError(f'Unknown CI environment {ci_env}')


@click.command(name='exec', context_settings={"ignore_unknown_options": True})
@click.pass_context
@click.argument('cmd_name', required=False, metavar='COMMAND')
@click.argument('extra_args', required=False, nargs=-1, metavar='EXTRA_ARGS')
@click.option('--command', help='Raw commandline to execute in docker container', metavar='\'<cmdline>\'')
@click.option('--save-as', default=None, help='Alias under which to save command specified with --command', metavar='<alias>', show_default=True)
@click.option('--warn/--no-warn', default=True, help='Whether to print a warning if overwriting an existing command.', show_default=True)
def exec_(ctx, cmd_name, extra_args, command, save_as, warn):
    """
        Execute COMMAND in a docker container. COMMAND is a named command which was previously saved using `--save-as`.
        If COMMAND is not provided, execute the command specified by --command. If --save-as is also provided,
        save the command under that name for later use with COMMAND. If EXTRA_ARGS are provided they will be appended
        to the command at runtime
    """
    if not cmd_name and not command:
        click.echo(exec_.get_help(ctx))
        ctx.exit(0)

    path, config = get_devenv_config()
    section = config[DEVENV_SECTION_NAME]

    if cmd_name and command:
        raise click.UsageError(
            'Cannot specify a command name and a command'
        )

    if cmd_name:
        key = f'cmd.{cmd_name}'
        if key not in section:
            raise click.UsageError(
                f'Command `{cmd_name}` not found',
            )
        cmd_to_run = section.get(key)
    else:
        cmd_to_run = command
        if save_as:
            add_devenv_command(section, save_as, cmd_to_run, warn)
            write_config_file(config, path)

    returncode = _run_command(section, path, cmd_to_run, extra_args, ctx.obj.debug)
    ctx.exit(returncode)



"""
    lager.lister.commands

    List commands
"""
import click
from texttable import Texttable
from .. import SUPPORTED_DEVICES
from ..config import read_config_file

@click.group(name='list')
def lister():
    """
        Lager list commands
    """
    pass


@lister.command()
@click.pass_context
def gateways(ctx):
    """
        List a user's gateways
    """

    session = ctx.obj.session
    resp = session.list_gateways()
    resp.raise_for_status()

    table = Texttable()
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(['t', 'i'])
    table.set_cols_align(["l", "r"])
    table.add_row(['name', 'id'])

    for gateway in resp.json()['gateways']:
        table.add_row([gateway['name'], gateway['id']])
    click.echo(table.draw())

@lister.command()
def supported_devices():
    """
        List supported devices
    """
    for device in SUPPORTED_DEVICES:
        click.echo(device)

@lister.command()
def secret_token():
    """
        Show the secret token, which can be used within a third-party CI system to auth lager-cli
    """
    config = read_config_file()
    click.echo(config['AUTH']['refresh'])



"""
    lager.erase.commands

    Commands for erasing a DUT
"""
import click
from ..context import get_default_gateway
from ..paramtypes import MemoryAddressType
from ..util import stream_output

@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.argument('start_addr', type=MemoryAddressType())
@click.argument('length', type=MemoryAddressType())
def erase(ctx, gateway, start_addr, length):
    """
        Erase DUT
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session
    addresses = dict(start_addr=start_addr, length=length)
    resp = session.erase_dut(gateway, addresses=addresses)
    stream_output(resp)



import click

def test_matcher_factory(test_runner):
    """
        Return matcher for named test_runner
    """
    if test_runner == 'unity':
        return UnityMatcher
    if test_runner is None or test_runner == 'none':
        return EmptyMatcher

    raise ValueError(f'Unknown test matcher {test_runner}')

def echo_line(line, color):
    """
        Try to echo a line with color, otherwise just emit it raw
    """
    try:
        decoded_line = line.decode()
        click.secho(decoded_line, fg=color)
    except UnicodeDecodeError:
        click.echo(line)

class UnityMatcher:
    summary_separator = b'-----------------------'

    def __init__(self, io):
        self.state = b''
        self.separator = None
        self.has_fail = False
        self.in_summary = False
        self.io = io

    def feed(self, data):
        self.state += data
        if b'\n' not in data:
            return

        lines = self.state.split(b'\n')
        to_process, remainder = lines[:-1], lines[-1]
        self.state = remainder
        for line in to_process:
            if line == self.summary_separator:
                self.in_summary = True
                click.echo(line)
                continue
            if self.in_summary:
                color = 'red' if self.has_fail else 'green'
                echo_line(line, color)
            else:
                if b':FAIL' in line:
                    self.has_fail = True
                    echo_line(line, 'red')
                elif b':PASS' in line:
                    echo_line(line, 'green')
                elif b':INFO' in line:
                    echo_line(line, 'yellow')
                else:
                    click.echo(line)

    def done(self):
        pass

    @property
    def exit_code(self):
        if self.has_fail:
            return 1
        return 0

class EmptyMatcher:
    def __init__(self, io):
        self.io = io

    def feed(self, data):
        self.io.output(data)

    def done(self):
        pass

    @property
    def exit_code(self):
        return 0



"""
    lager.auth.commands

    auth commands
"""

import time
import urllib.parse
import webbrowser
import requests
import click
from . import (
    get_client_id, get_auth_url, get_audience,
    read_config_file, write_config_file,
)
from ..context import LagerContext

SCOPE = 'read:gateway flash:duck offline_access'

def poll_for_token(device_code, interval):
    """
        Poll for an auth token for the specified device at the given interval
    """
    data = {
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
        'device_code': device_code,
        'client_id': get_client_id(),
    }
    token_url = urllib.parse.urljoin(get_auth_url(), '/oauth/token')
    while True:
        resp = requests.post(token_url, data=data)
        if resp.status_code == 200:
            return resp.json()

        if resp.status_code >= 500:
            resp.raise_for_status()

        error = resp.json()['error']
        if error == 'authorization_pending':
            time.sleep(interval)
        elif error == 'expired_token':
            click.secho('Session timed out. Please run `lager login` again.', err=True, fg='red')
            click.get_current_context().exit(1)
        elif error == 'access_denied':
            click.secho('Access denied.', err=True, fg='red')
            click.get_current_context().exit(1)


@click.command()
@click.pass_context
def login(ctx):
    """
        Log in
    """
    has_browser = True
    try:
        webbrowser.get()
    except webbrowser.Error:
        has_browser = False

    data = {
        'audience': get_audience(),
        'scope': SCOPE,
        'client_id': get_client_id(),
    }
    code_url = urllib.parse.urljoin(get_auth_url(), '/oauth/device/code')
    response = requests.post(code_url, data=data)
    response.raise_for_status()

    code_response = response.json()
    uri = code_response['verification_uri_complete']
    user_code = code_response['user_code']
    if has_browser:
        click.echo('Please confirm the following code appears in your browser: ', nl=False)
        click.secho(user_code, fg='green')
        if click.confirm('Lager would like to open a browser window to confirm your login info'):
            webbrowser.open_new(uri)
        else:
            click.echo('Cancelled')
    else:
        click.echo('Please visit ', nl=False)
        click.secho(uri, fg='green', nl=False)
        click.echo(' in your browser')
        click.echo('And confirm your device token: ', nl=False)
        click.secho(user_code, fg='green')

    click.echo('Awaiting confirmation... (Could take up to 5 seconds after clicking "Confirm" in your browser)')
    payload = poll_for_token(code_response['device_code'], code_response['interval'])

    config = read_config_file()
    config['AUTH'] = {
        'token': payload['access_token'],
        'type': payload['token_type'],
        'refresh': payload['refresh_token'],
    }
    write_config_file(config)

    ctx = LagerContext(
        ctx=ctx,
        auth=config['AUTH'],
        defaults=None,
        debug=False,
        style=None
    )

    session = ctx.session
    url = 'cli/post-login'
    session.post(url)

    click.secho('Success! You\'re ready to use Lager!', fg='green')

@click.command()
def logout():
    """
        Log out
    """
    try:
        config = read_config_file()
    except FileNotFoundError:
        return

    if 'AUTH' in config:
        del config['AUTH']

    write_config_file(config)



"""
    lager.connect.commands

    Commands for connecting to / disconnecting from DUT
"""
import itertools
import click
from .. import SUPPORTED_DEVICES, SUPPORTED_INTERFACES
from ..context import get_default_gateway
from ..paramtypes import HexParamType, VarAssignmentType

@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.option(
    '--snr',
    help='Serial number of device to connect. Required if multiple DUTs connected to gateway')
@click.option('--device', help='Target device type', type=click.Choice(SUPPORTED_DEVICES), required=True)
@click.option('--interface', help='Target interface', type=click.Choice(SUPPORTED_INTERFACES), default='ftdi', show_default=True)
@click.option('--transport', help='Target transport', type=click.Choice(['swd', 'jtag']), default='swd', show_default=True)
@click.option('--speed', help='Target interface speed in kHz', required=False, default='adaptive', show_default=True)
@click.option('--workareasize', help='Set work area size. Useful for STM32 chips.', type=HexParamType(), required=False, default=None)
@click.option('--set', 'set_', multiple=True, type=VarAssignmentType(), help='Set debugger environment variable FOO to BAR')
@click.option('--force', is_flag=True, default=False, help='Disconnect debugger before reconnecting. If not set, connect will fail if debugger is already connected. Cannot be used with --ignore-if-connected', show_default=True)
@click.option('--ignore-if-connected', is_flag=True, default=False, help='If debugger is already connected, skip connection attempt and exit with success. Cannot be used with --force', show_default=True)
def connect(ctx, gateway, snr, device, interface, transport, speed, workareasize, set_, force, ignore_if_connected):
    """
        Connect your gateway to your Device Under Test (DUT).

        Examples
        --------
        Connect to an nrf52
           `lager connect \\-\\-device nrf52`
    """
    if force and ignore_if_connected:
        click.secho('Cannot specify --force and --ignore-if-connected', fg='red')
        ctx.exit(1)

    set_ = list(set_)
    if workareasize:
        set_.append(['WORKAREASIZE', hex(workareasize)])
    if gateway is None:
        gateway = get_default_gateway(ctx)

    # Step one, try to start gdb on gateway
    files = []
    if snr:
        files.append(('snr', snr))
    files.append(('device', device))
    files.append(('interface', interface))
    files.append(('transport', transport))
    files.append(('speed', speed))
    files.append(('force', force))
    files.append(('ignore_if_connected', ignore_if_connected))
    files.extend(
        zip(itertools.repeat('varnames'), [name for (name, _value) in set_])
    )
    files.extend(
        zip(itertools.repeat('varvalues'), [value for (_name, value) in set_])
    )

    session = ctx.obj.session
    resp = session.start_debugger(gateway, files=files).json()
    if resp.get('start') == 'ok':
        click.secho('Connected!', fg='green')
    elif resp.get('already_running') == 'ok':
        click.secho('Debugger already connected, ignoring', fg='green')

@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
def disconnect(ctx, gateway):
    """
        Disconnect your gateway from your Device Under Test (DUT).
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session
    session.stop_debugger(gateway)
    click.secho('Disconnected!', fg='green')



"""
    lager.gateway.tunnel

    GDB tunnel functions
"""
import functools
import logging
import click
import trio
import lager_trio_websocket as trio_websocket
from ..util import heartbeat

logger = logging.getLogger(__name__)

async def send_to_websocket(websocket, gdb_client_stream, nursery):
    """
        Read data from gdb_client_stream (a trio stream connected to a gdb client)
        and send to websocket (ultimate destination is gateway gdbserver).
    """
    try:
        async with gdb_client_stream:
            async for msg in gdb_client_stream:
                await websocket.send_message(msg)
    except trio.BrokenResourceError:
        pass
    finally:
        nursery.cancel_scope.cancel()


async def send_to_gdb(websocket, gdb_client_stream, nursery):
    """
        Read data from websocket (originating from gateway gdbserver)
        and send to gdb_client_stream (a trio stream connected to a gdb client)
    """
    while True:
        try:
            msg = await websocket.get_message()
            await gdb_client_stream.send_all(msg)
        except trio_websocket.ConnectionClosed:
            nursery.cancel_scope.cancel()

async def cloud_connection_handler(connection_params, gdb_client_stream):
    """
        Handle a single connection from a gdb client
    """
    (uri, kwargs) = connection_params
    sockname = gdb_client_stream.socket.getsockname()
    click.echo(f'Serving gdb client: {sockname}')
    try:
        async with trio_websocket.open_websocket_url(uri, disconnect_timeout=1, **kwargs) as websocket:
            async with trio.open_nursery() as nursery:
                nursery.start_soon(send_to_websocket, websocket, gdb_client_stream, nursery)
                nursery.start_soon(send_to_gdb, websocket, gdb_client_stream, nursery)
                nursery.start_soon(heartbeat, websocket, 30, 30)
    except Exception as exc:  # pylint: disable=broad-except
        logger.exception('Exception in connection_handler', exc_info=exc)
    finally:
        click.echo(f'gdb client disconnected: {sockname}')

async def send_to_local_client(gateway_stream, gdb_client_stream, nursery):
    try:
        async for data in gateway_stream:
            await gdb_client_stream.send_all(data)
        try:
            await gdb_client_stream.send_eof()
        except trio.BrokenResourceError:
            pass
    except Exception as exc:
        logger.exception('send_to_local_client failed: ', exc_info=exc)
        nursery.cancel_scope.cancel()

async def send_to_local_gateway(gateway_stream, gdb_client_stream, nursery):
    try:
        async for data in gdb_client_stream:
            await gateway_stream.send_all(data)
        try:
            await gateway_stream.send_eof()
        except trio.BrokenResourceError:
            pass
    except Exception as exc:
        logger.exception('send_to_local_gateway failed: ', exc_info=exc)
        nursery.cancel_scope.cancel()

async def local_connection_handler(session, gateway, remote_host, remote_port, gdb_client_stream):
    """
        Handle a single connection from a gdb client
    """
    sockname = gdb_client_stream.socket.getsockname()
    if not remote_host:
        resp = await trio.to_thread.run_sync(session.start_local_gdb_tunnel, gateway, False)
        resp = resp.json()
        # resp = session.start_local_gdb_tunnel(gateway, False).json()
        remote_host = resp['host']
        remote_port = int(resp['port'])
    click.echo(f'Serving gdb client: {sockname}')
    try:
        async with await trio.open_tcp_stream(remote_host, remote_port) as gateway_stream:
            async with trio.open_nursery() as nursery:
                nursery.start_soon(send_to_local_client, gateway_stream, gdb_client_stream, nursery)
                nursery.start_soon(send_to_local_gateway, gateway_stream, gdb_client_stream, nursery)
    except Exception as exc:  # pylint: disable=broad-except
        logger.exception('Exception in connection_handler', exc_info=exc)
    finally:
        click.echo(f'gdb client disconnected: {sockname}')

async def serve_tunnel(host, port, connection_params, *, task_status=trio.TASK_STATUS_IGNORED):
    """
        Start up the server that tunnels traffic to a gdbserver instance running on a gateway
    """
    async with trio.open_nursery() as nursery:
        handler = functools.partial(cloud_connection_handler, connection_params)
        serve_listeners = functools.partial(trio.serve_tcp, handler, port, host=host)

        server = await nursery.start(serve_listeners)
        task_status.started(server)
        click.echo(f'Serving GDB on {host}:{port}. Press Ctrl+C to quit.')
        try:
            await trio.sleep_forever()
        except KeyboardInterrupt:
            nursery.cancel_scope.cancel()

async def serve_local_tunnel(session, gateway, host, port, fork, *, task_status=trio.TASK_STATUS_IGNORED):
    """
        Start up the server that locally tunnels traffic to a gdbserver instance running on a gateway
    """
    remote_host = None
    remote_port = None
    if fork:
        resp = session.start_local_gdb_tunnel(gateway, True).json()
        remote_host = resp['host']
        remote_port = int(resp['port'])

    async with trio.open_nursery() as nursery:
        handler = functools.partial(local_connection_handler, session, gateway, remote_host, remote_port)
        serve_listeners = functools.partial(trio.serve_tcp, handler, port, host=host)

        server = await nursery.start(serve_listeners)
        task_status.started(server)
        click.echo(f'Serving GDB on {host}:{port}. Press Ctrl+C to quit.')
        try:
            await trio.sleep_forever()
        except KeyboardInterrupt:
            nursery.cancel_scope.cancel()



"""
    lager.run.commands

    Commands for running a an image on a DUT
"""
import click
from ..context import get_default_gateway

@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
def run(ctx, gateway):
    """
        Run a DUT connected to a gateway
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    session = ctx.obj.session
    resp = session.run_dut(gateway)
    for chunk in resp.iter_content(chunk_size=1):
        click.echo(chunk, nl=False)



"""
    lager.testrun.commands

    Commands for flashing an image to a DUT and collecting results
"""
import math
import click
from ..context import get_default_gateway
from ..reset.commands import do_reset
from ..uart.commands import do_uart
from ..flash.commands import do_flash
from ..paramtypes import BinfileType
from ..util import stream_output
from ..status import run_job_output

@click.command()
@click.pass_context
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.option('--serial-device', help='Gateway serial port device path', metavar='path')
@click.option('--baudrate', help='Serial baud rate', type=int, default=None)
@click.option('--bytesize', help='Number of data bits', type=click.Choice(['5', '6', '7', '8']), default=None)
@click.option('--parity', help='Parity check', type=click.Choice(['none', 'even', 'odd', 'mark', 'space']), default=None)
@click.option('--stopbits', help='Number of stop bits', type=click.Choice(['1', '1.5', '2']), default=None)
@click.option('--xonxoff/--no-xonxoff', default=None, help='Enable/disable software XON/XOFF flow control')
@click.option('--rtscts/--no-rtscts', default=None, help='Enable/disable hardware RTS/CTS flow control')
@click.option('--dsrdtr/--no-dsrdtr', default=None, help='Enable/disable hardware DSR/DTR flow control')
@click.option('--test-runner', help='End the UART session when end-of-test is detected', type=click.Choice(['none', 'unity']), default='unity')
@click.option('--interactive', is_flag=True, help='Run as an interactive TTY session', default=False)
@click.option('--message-timeout', help='Message timeout', type=click.FLOAT, default=math.inf)
@click.option('--overall-timeout', help='Overall timeout', type=click.FLOAT, default=math.inf)
@click.option(
    '--hexfile',
    multiple=True, type=click.Path(exists=True),
    help='Hexfile(s) to flash. May be passed multiple times; files will be flashed in order.')
@click.option(
    '--binfile',
    multiple=True, type=BinfileType(exists=True),
    help='Binfile(s) to flash. Syntax: --binfile `<filename>,<address>` '
         'May be passed multiple times; files will be flashed in order.')
@click.option(
    '--preverify/--no-preverify',
    help='If true, only flash target if image differs from current flash contents',
    default=True)
@click.option('--verify/--no-verify', help='Verify image successfully flashed', default=True)
@click.option('--display-job-id', default=False, is_flag=True)
def testrun(ctx, gateway, serial_device, baudrate, bytesize, parity, stopbits, xonxoff, rtscts,
            dsrdtr, test_runner, interactive, message_timeout, overall_timeout, hexfile, binfile,
            preverify, verify, display_job_id):
    """
        Flash and run test on a DUT connected to a gateway
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)
    session = ctx.obj.session
    resp = do_reset(session, gateway, halt=True)
    stream_output(resp)

    resp = do_uart(
        ctx, gateway, serial_device, baudrate, bytesize, parity,
        stopbits, xonxoff, rtscts, dsrdtr, test_runner
    )
    test_run = resp.json()

    resp = do_flash(session, gateway, hexfile, binfile, preverify, verify)
    stream_output(resp)

    resp = do_reset(ctx.obj.session, gateway, halt=False)
    stream_output(resp)

    job_id = test_run['test_run']['id']
    if display_job_id:
        click.echo('Job id: {}'.format(job_id), err=True)

    connection_params = ctx.obj.websocket_connection_params(socktype='job', job_id=job_id)
    run_job_output(
        connection_params, test_runner, interactive, message_timeout,
        overall_timeout, None, ctx.obj.debug,
    )



"""
Lager CLI
----
A Command Line Interface for Lager Data
"""

__version__ = '0.1.29'

SUPPORTED_DEVICES = (
    'nrf52',
    'cc3220sf',
    'cc3235sf',
    'cc3220s',
    'atsame70',
    'efm32',
    'stm32f0x',
    'stm32f1x',
    'stm32f2x',
    'stm32f3x',
    'stm32f4x',
    'stm32f7x',
    'stm32g0x',
    'stm32g4x',
    'stm32h7x',
    'stm32h7x_dual_bank',
    'stm32l0',
    'stm32l0_dual_bank',
    'stm32l1',
    'stm32l1x_dual_bank',
    'stm32l4x',
    'stm32mp15x',
    'stm32w108xx',
    'stm32wbx',
    'stm32wlx',
    'stm32xl',
)

SUPPORTED_INTERFACES = (
    'ftdi',
    'jlink',
    'cmsis-dap',
    'xds110',
    'stlink',
    'stlink-dap',
)



"""
    lager.config

    Config file management routines
"""
import os
import configparser
import click

DEFAULT_CONFIG_FILE_NAME = '.lager'
_LAGER_CONFIG_FILE_NAME = os.getenv('LAGER_CONFIG_FILE_NAME', DEFAULT_CONFIG_FILE_NAME)

DEVENV_SECTION_NAME = 'DEVENV'


def _get_global_config_file_path():
    return make_config_path(os.path.expanduser('~'))

def make_config_path(directory, config_file_name=None):
    """
        Make a full path to a lager config file
    """
    if config_file_name is None:
        config_file_name = _LAGER_CONFIG_FILE_NAME

    return os.path.join(directory, config_file_name)

def find_devenv_config_path():
    """
        Find a local .lager config, if it exists
    """
    configs = _find_config_files()
    if not configs:
        return None
    return configs[0]

def _find_config_files():
    """
        Search up from current directory for all .lager files
    """
    cwd = os.getcwd()
    cfgs = []
    global_config_file_path = _get_global_config_file_path()
    while True:
        config_path = make_config_path(cwd)
        if os.path.exists(config_path) and config_path != global_config_file_path:
            cfgs.append(config_path)
        parent = os.path.dirname(cwd)
        if parent == cwd:
            break
        cwd = parent

    return cfgs


def read_config_file(path=None):
    """
        Read our config file into `config` object
    """
    if path is None:
        path = _get_global_config_file_path()
    config = configparser.SafeConfigParser()
    try:
        with open(path) as f:
            config.read_file(f)
    except FileNotFoundError:
        pass

    if 'LAGER' not in config:
        config.add_section('LAGER')
    return config

def write_config_file(config, path=None):
    """
        Write out `config` into our config file
    """
    if path is None:
        path = _get_global_config_file_path()
    with open(path, 'w') as f:
        config.write(f)

def add_devenv_command(section, command_name, command, warn):
    """
        Add a named command to devenv
    """
    key = f'cmd.{command_name}'
    if key in section and warn:
        click.echo(f'Command `{command_name}` already exists, overwriting. ', nl=False, err=True)
        click.echo(f'Previous value: {section[key]}', err=True)
    section[key] = command

def remove_devenv_command(section, command_name):
    """
        Delete a named command
    """
    key = f'cmd.{command_name}'
    if key not in section:
        click.secho(f'Command `{command_name}` does not exist.', fg='red', err=True)
        click.get_current_context().exit(1)
    del section[key]

def all_commands(section):
    """
        Get a map of command name -> command for all commands in the section
    """
    return {
        k.split('.', 1)[1]: section[k] for k in section.keys() if k.startswith('cmd.')
    }

def get_devenv_config():
    """
        Return a path and config file for devenv
    """
    config_path = find_devenv_config_path()
    if config_path is None:
        click.echo(f'Could not find {_LAGER_CONFIG_FILE_NAME} in {os.getcwd()} or any parent directories', err=True)
        click.get_current_context().exit(1)
    config = read_config_file(config_path)
    return config_path, config



"""
    lager.status.commands

    Status commands
"""
import math
import click
from ..status import run_job_output

@click.group(name='job')
def job():
    """
        Lager job commands
    """
    pass

@job.command()
@click.pass_context
@click.argument('job_id')
@click.option('--message-timeout', default=math.inf, type=click.FLOAT,
              help='Max time in seconds to wait between messages from API.'
              'This timeout only affects reading output and does not cancel the actual test run if hit.')
@click.option('--overall-timeout', default=math.inf, type=click.FLOAT,
              help='Cumulative time in seconds to wait for session output.'
              'This timeout only affects reading output and does not cancel the actual test run if hit.')
def status(ctx, job_id, message_timeout, overall_timeout):
    """
        Get job status
    """
    connection_params = ctx.obj.websocket_connection_params(socktype='job', job_id=job_id)
    run_job_output(connection_params, None, False, message_timeout, overall_timeout, 0, ctx.obj.debug)



"""
    lager.gpio.commands

    Commands for manipulating gateway GPIO lines
"""
import click
from ..context import get_default_gateway

@click.group()
def gpio():
    """
        Lager gpio commands
    """
    pass

_GPIO_CHOICES = click.IntRange(0, 3)
_LEVEL_CHOICES = click.Choice(('LOW', 'HIGH'), case_sensitive=False)

@gpio.command(name='set')
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.argument('gpio_', metavar='GPIO', type=_GPIO_CHOICES)
@click.argument('type_', type=click.Choice(('IN', 'OUT'), case_sensitive=False))
@click.option('--pull', type=click.Choice(('UP', 'DOWN', 'OFF'), case_sensitive=False), default='OFF', show_default=True, help='Sets or clears the internal GPIO pull-up/down resistor.')
@click.pass_context
def set_(ctx, gateway, gpio_, type_, pull):
    """
        Sets pin GPIO mode (input/output)

        If type is IN, --pull controls the internal GPIO pull-up/down resistor.
        Otherwise it has no effect.

        GPIO can be 0, 1, 2, or 3
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)

    if type_ == 'OUT' and pull != 'OFF':
        click.echo(f'GPIO pin {gpio_} set as output, ignoring --pull', err=True)

    ctx.obj.session.gpio_set(gateway, gpio_, type_, pull)


@gpio.command(name='input')
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.argument('gpio_', metavar='GPIO', type=_GPIO_CHOICES)
@click.pass_context
def input_(ctx, gateway, gpio_):
    """
        Returns GPIO level (0 for low, 1 for high)

        GPIO can be 0, 1, 2, or 3
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)
    result = ctx.obj.session.gpio_input(gateway, gpio_)
    click.echo(result.json()['level'])


@gpio.command()
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.argument('gpio_', metavar='GPIO', type=_GPIO_CHOICES)
@click.argument('level', type=_LEVEL_CHOICES)
@click.pass_context
def output(ctx, gateway, gpio_, level):
    """
        Sets GPIO level.

        GPIO can be 0, 1, 2, or 3
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)
    ctx.obj.session.gpio_output(gateway, gpio_, level)

@gpio.command()
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.argument('gpio_', metavar='GPIO', type=_GPIO_CHOICES)
@click.option('--pulsewidth', type=click.IntRange(500, 2500), help='500 (most anti-clockwise) - 2500 (most clockwise).')
@click.option('--stop', is_flag=True, default=False, help='Stops servo pulses on the GPIO')
@click.pass_context
def servo(ctx, gateway, gpio_, pulsewidth, stop):
    """
        Starts (--pulsewidth 500-2500) or stops (--stop) servo pulses on the GPIO.
        The pulsewidths supported by servos varies and should probably be determined by experiment.
        A value of 1500 should always be safe and represents the mid-point of rotation.
        You can DAMAGE a servo if you command it to move beyond its limits.

        GPIO can be 0, 1, 2, or 3
    """
    if stop and pulsewidth:
        click.echo(ctx.get_help(), err=True)
        click.secho('--stop and --pulsewidth both specified; please only use one', fg='red', err=True)
        ctx.exit(1)
    if not stop and not pulsewidth:
        click.echo(ctx.get_help(), err=True)
        click.secho('Please pass one of --pulsewidth or --stop', fg='red', err=True)
        ctx.exit(1)
    if gateway is None:
        gateway = get_default_gateway(ctx)
    ctx.obj.session.gpio_servo(gateway, gpio_, pulsewidth, stop)

@gpio.command()
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.argument('gpio_', metavar='GPIO', type=_GPIO_CHOICES)
@click.option('--pulse-length', type=click.IntRange(1, 100), help='Pulse length in microseconds (1-100)', required=True)
@click.option('--level', type=_LEVEL_CHOICES, help='Pulse level', required=True)
@click.pass_context
def trigger(ctx, gateway, gpio_, pulse_length, level):
    """
        Send a trigger pulse to GPIO.
        The GPIO is set to level for pulse-length microseconds and then reset to not level.

        GPIO can be 0, 1, 2, or 3
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)
    ctx.obj.session.gpio_trigger(gateway, gpio_, pulse_length, level)

@gpio.command()
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.option('--frequency', type=click.IntRange(0, 187_500_000), help='Frequency in Hz (1-187.5M or 0 for off)', required=True)
@click.option('--dutycycle', type=click.IntRange(0, 1_000_000), help='Duty cycle. 0 (off) to 1,000,000 (fully on).', required=True)
@click.pass_context
def hardware_pwm(ctx, gateway, frequency, dutycycle):
    """
        Starts hardware PWM on GPIO lines 1 and 3 at the specified frequency and dutycycle.

        Frequencies above 30MHz are unlikely to work.

        The actual number of steps beween off and fully on is the integral part of 375M/frequency

        The actual frequency set is 375M/steps

        There will only be a million steps for a frequency of 375.
        Lower frequencies will have more steps and higher frequencies will have fewer steps.
        --dutycycle is automatically scaled to take this into account.
    """
    if gateway is None:
        gateway = get_default_gateway(ctx)
    ctx.obj.session.gpio_hardware_pwm(gateway, frequency, dutycycle)


@gpio.command()
@click.option('--gateway', required=False, help='ID of gateway to which DUT is connected')
@click.argument('frequency', type=click.INT, required=True)
@click.pass_context
def hardware_clock(ctx, gateway, frequency):
    """
        Starts hardware clock on GPIO line 0 (PWM with 50% duty cycle)

        FREQUENCY can be 0 for off, or an integer between 13,184 and 375,000,000

        Frequencies above 30MHz are unlikely to work.
    """
    if frequency != 0 and (frequency < 13184 or frequency > 375_000_000):
        click.echo('Frequency must be 0 (off) or between 13,184 - 375,000,000')
        ctx.exit(1)
    if gateway is None:
        gateway = get_default_gateway(ctx)
    ctx.obj.session.gpio_hardware_clock(gateway, frequency)




import os
import setuptools

from lager_cli import __version__ as lager_version

def readme():
    path = os.path.dirname(__file__)
    with open(os.path.join(path, 'README.md')) as f:
        return f.read()

name = 'lager-cli'
description = 'Lager Command Line Interface'
author = 'Lager Data LLC'
email = 'hello@lagerdata.com'
classifiers = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Topic :: Software Development',
]

if __name__ == "__main__":
    setuptools.setup(
        name=name,
        version=lager_version,
        description=description,
        long_description=readme(),
        classifiers=classifiers,
        url='https://github.com/lagerdata/lager-cli',
        author=author,
        author_email=email,
        maintainer=author,
        maintainer_email=email,
        license='MIT',
        python_requires=">=3.6",
        packages=setuptools.find_packages(),
        install_requires='''
            async-generator == 1.10
            bson == 0.5.10
            certifi == 2020.4.5.1
            chardet == 3.0.4
            click == 7.1.2
            colorama == 0.4.3
            h11 == 0.9.0
            idna == 2.9
            ipaddress == 1.0.23
            multidict == 4.7.6
            outcome == 1.0.1
            python-dateutil == 2.8.1
            requests == 2.23.0
            requests-toolbelt == 0.9.1
            six == 1.15.0
            sniffio == 1.1.0
            sortedcontainers == 2.2.2
            tenacity == 6.2.0
            texttable == 1.6.2
            trio == 0.16.0
            lager-trio-websocket == 0.9.0-dev
            urllib3 == 1.25.9
            wsproto == 0.14.1
            yarl == 1.4.2
        ''',
        entry_points={
            'console_scripts': [
                'lager=lager_cli.cli:cli',
            ],
        }
    )



"""
    lager.cli

    Command line interface entry point
"""
import os
import urllib.parse

import traceback
import click

from . import __version__
from .config import read_config_file
from .context import LagerContext

from .gateway.commands import _gateway
from .setter.commands import setter
from .lister.commands import lister
from .auth import load_auth
from .auth.commands import login, logout
from .job.commands import job
from .devenv.commands import devenv
from .exec.commands import exec_
from .flash.commands import flash
from .run.commands import run
from .erase.commands import erase
from .reset.commands import reset
from .uart.commands import uart
from .testrun.commands import testrun
from .gdbserver.commands import gdbserver
from .connect.commands import connect, disconnect
from .gpio.commands import gpio

def _decode_environment():
    for key in os.environ:
        if key.startswith('LAGER_'):
            os.environ[key] = urllib.parse.unquote(os.environ[key])

@click.group(invoke_without_command=True)
@click.pass_context
@click.option('--version', 'see_version', is_flag=True, help='See package version')
@click.option('--debug', 'debug', is_flag=True, help='Show debug output', default=False)
@click.option('--colorize', 'colorize', is_flag=True, help='Color output', default=False)
def cli(ctx=None, see_version=None, debug=False, colorize=False):
    """
        Lager CLI
    """
    if os.getenv('LAGER_DECODE_ENV'):
        _decode_environment()

    if see_version:
        click.echo(__version__)
        click.get_current_context().exit(0)
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
    else:
        os_args = click.get_os_args()
        help_invoked = '--help' in os_args
        skip_auth = ctx.invoked_subcommand in ('login', 'logout', 'set', 'devenv', 'exec') or help_invoked
        setup_context(ctx, debug, colorize, skip_auth)

cli.add_command(_gateway)
cli.add_command(setter)
cli.add_command(lister)
cli.add_command(login)
cli.add_command(logout)
cli.add_command(job)
cli.add_command(devenv)
cli.add_command(exec_)
cli.add_command(flash)
cli.add_command(run)
cli.add_command(erase)
cli.add_command(reset)
cli.add_command(uart)
cli.add_command(testrun)
cli.add_command(gdbserver)
cli.add_command(connect)
cli.add_command(disconnect)
cli.add_command(gpio)

def setup_context(ctx, debug, colorize, skip_auth):
    """
        Ensure the user has a valid authorization
    """
    auth = None
    if not skip_auth:
        try:
            auth = load_auth()
        except Exception:  # pylint: disable=broad-except
            trace = traceback.format_exc()
            click.secho(trace, fg='red')
            click.echo('Something went wrong. Please run `lager logout` followed by `lager login`')
            click.echo('For additional assistance please send the above traceback (in red) to support@lagerdata.com')
            click.get_current_context().exit(0)

        if not auth:
            click.echo('Please login using `lager login` first')
            click.get_current_context().exit(0)

    config = read_config_file()
    ctx.obj = LagerContext(
        ctx=ctx,
        auth=auth,
        defaults=config['LAGER'],
        debug=debug,
        style=click.style if colorize else lambda string, **kwargs: string,
    )



"""
    lager.devenv.commands

    Devenv commands
"""
import os
import subprocess
from pathlib import Path
import click
from ..config import (
    read_config_file,
    write_config_file,
    add_devenv_command,
    remove_devenv_command,
    all_commands,
    find_devenv_config_path,
    make_config_path,
    get_devenv_config,
    DEVENV_SECTION_NAME,
)

@click.group()
def devenv():
    """
        Lager devenv commands
    """
    pass

existing_dir_type = click.Path(
    exists=True,
    file_okay=False,
    dir_okay=True,
    readable=True,
    resolve_path=True,
)

@devenv.command()
@click.pass_context
@click.option('--image', prompt='Docker image', default='lagerdata/devenv-cortexm', show_default=True)
@click.option('--mount-dir', prompt='Source code mount directory in docker container',
              default='/app', show_default=True)
@click.option('--shell', prompt='Path to shell executable in docker image',
              default='/bin/bash', show_default=True)
def create(ctx, image, mount_dir, shell):
    """
        Create a development environment
    """
    config_path = find_devenv_config_path()
    if config_path is not None:
        answer = click.confirm('Config file {} exists; overwrite?'.format(config_path))
        if not answer:
            ctx.exit(0)

    if config_path is None:
        config_path = make_config_path(os.getcwd())
        Path(config_path).touch()

    config = read_config_file(config_path)
    if not config.has_section(DEVENV_SECTION_NAME):
        config.add_section(DEVENV_SECTION_NAME)
    config.set(DEVENV_SECTION_NAME, 'image', image)
    config.set(DEVENV_SECTION_NAME, 'mount_dir', mount_dir)
    config.set(DEVENV_SECTION_NAME, 'shell', shell)
    write_config_file(config, config_path)

@devenv.command()
def terminal():
    """
        Start an interactive terminal for a docker image
    """
    path, config = get_devenv_config()
    section = config[DEVENV_SECTION_NAME]

    image = section.get('image')
    source_dir = os.path.dirname(path)
    mount_dir = section.get('mount_dir')
    subprocess.run([
        'docker',
        'run',
        '-it',
        '--rm',
        '-v',
        f'{source_dir}:{mount_dir}',
        image,
    ], check=True)


@devenv.command()
@click.confirmation_option(prompt='Are you sure you want to delete your devenv?')
def delete():
    """
        Delete devenv config
    """
    config_path = find_devenv_config_path()
    if not config_path or not os.path.exists(config_path):
        return

    os.remove(config_path)

@devenv.command()
@click.argument('command_name')
@click.argument('command', required=False)
@click.option('--warn/--no-warn', default=True, help='Whether to print a warning if overwriting an existing command.', show_default=True)
def add_command(command_name, command, warn):
    """
        Add COMMAND to devenv with the name COMMAND_NAME
    """
    path, config = get_devenv_config()
    section = config[DEVENV_SECTION_NAME]
    if not command:
        command = click.prompt('Please enter the command')

    add_devenv_command(section, command_name, command, warn)
    write_config_file(config, path)

@devenv.command()
@click.argument('command_name')
@click.option('--devenv', '_devenv', help='Delete command from devenv named `foo`', metavar='foo')
def delete_command(command_name, _devenv):
    """
        Delete COMMAND_NAME from devenv
    """
    path, config = get_devenv_config()
    section = config[DEVENV_SECTION_NAME]

    remove_devenv_command(section, command_name)
    write_config_file(config, path)


@devenv.command()
def commands():
    """
        List the commands in a devenv
    """
    _, config = get_devenv_config()
    section = config[DEVENV_SECTION_NAME]
    for name, command in all_commands(section).items():
        click.secho(name, fg='green', nl=False)
        click.echo(f': {command}')



"""
    lager.context

    CLI context management
"""
from enum import Enum
import functools
import os
import json
import ssl
import urllib.parse
import urllib3
import requests
import click
from requests_toolbelt.sessions import BaseUrlSession

_DEFAULT_HOST = 'https://app.lagerdata.com'
_DEFAULT_WEBSOCKET_HOST = 'wss://app.lagerdata.com'

def print_openocd_error(error):
    """
        Parse an openocd log file and print the error lines
    """
    if not error:
        return
    parsed = json.loads(error)
    logfile = parsed['logfile']
    if not logfile:
        return
    for line in logfile.splitlines():
        if 'Error: ' in line:
            click.secho(line, fg='red', err=True)

OPENOCD_ERROR_CODES = set((
    'openocd_start_failed',
))

def quote(gateway):
    return urllib.parse.quote(str(gateway), safe='')

class LagerSession(BaseUrlSession):
    """
        requests session wrapper
    """

    @staticmethod
    def handle_errors(ctx, r, *args, **kwargs):
        """
            Handle request errors
        """
        try:
            current_context = click.get_current_context()
            ctx = current_context
        except RuntimeError:
            pass
        if r.status_code == 404:
            name = ctx.params['gateway'] or ctx.obj.default_gateway
            click.secho('You don\'t have a gateway with id `{}`'.format(name), fg='red', err=True)
            click.secho(
                'Please double check your login credentials and gateway id',
                fg='red',
                err=True,
            )
            ctx.exit(1)
        if r.status_code == 422:
            error = r.json()['error']
            if error['code'] in OPENOCD_ERROR_CODES:
                print_openocd_error(error['description'])
            else:
                click.secho(error['description'], fg='red', err=True)
            ctx.exit(1)
        if r.status_code >= 500:
            click.secho('Something went wrong with the Lager API', fg='red', err=True)
            ctx.exit(1)

        r.raise_for_status()

    def __init__(self, ctx, auth, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connection_exception = None
        verify = 'NOVERIFY' not in os.environ
        if not verify:
            urllib3.disable_warnings()

        if auth:
            auth_header = {
                'Authorization': '{} {}'.format(auth['type'], auth['token'])
            }
            self.headers.update(auth_header)
        self.verify = verify
        self.hooks['response'].append(functools.partial(LagerSession.handle_errors, ctx))


    def request(self, *args, **kwargs):
        """
            Catch connection errors so they can be handled more cleanly
        """
        try:
            return super().request(*args, **kwargs)
        except requests.exceptions.ConnectTimeout:
            click.secho('Connection to Lager API timed out', fg='red', err=True)
            click.get_current_context().exit(1)
        except requests.exceptions.ConnectionError:
            click.secho('Could not connect to Lager API', fg='red', err=True)
            click.get_current_context().exit(1)

    def start_debugger(self, gateway, files):
        """
            Start the debugger on the gateway
        """
        url = 'gateway/{}/start-debugger'.format(quote(gateway))
        return self.post(url, files=files)

    def stop_debugger(self, gateway):
        """
            Stop the debugger on the gateway
        """
        url = 'gateway/{}/stop-debugger'.format(quote(gateway))
        return self.post(url)

    def erase_dut(self, gateway, addresses):
        """
            Erase DUT connected to gateway
        """
        url = 'gateway/{}/erase-duck'.format(quote(gateway))
        return self.post(url, json=addresses)

    def flash_dut(self, gateway, files):
        """
            Flash DUT connected to gateway
        """
        url = 'gateway/{}/flash-duck'.format(quote(gateway))
        return self.post(url, files=files, stream=True)

    def gateway_hello(self, gateway):
        """
            Say hello to gateway to see if it is connected
        """
        url = 'gateway/{}/hello'.format(quote(gateway))
        return self.get(url)

    def serial_numbers(self, gateway, model):
        """
            Get serial numbers of devices attached to gateway
        """
        url = 'gateway/{}/serial-numbers'.format(quote(gateway))
        return self.get(url, params={'model': model})

    def serial_ports(self, gateway):
        """
            Get serial port devices attached to gateway
        """
        url = 'gateway/{}/serial-ports'.format(quote(gateway))
        return self.get(url)

    def gateway_status(self, gateway):
        """
            Get debugger status on gateway
        """
        url = 'gateway/{}/status'.format(quote(gateway))
        return self.get(url)

    def list_gateways(self):
        """
            Get all gateways for logged-in user
        """
        url = 'gateway/list'
        return self.get(url)

    def reset_dut(self, gateway, halt):
        """
            Reset the DUT attached to a gateway and optionally halt it
        """
        url = 'gateway/{}/reset-duck'.format(quote(gateway))
        return self.post(url, json={'halt': halt})

    def run_dut(self, gateway):
        """
            Run the DUT attached to a gateway
        """
        url = 'gateway/{}/run-duck'.format(quote(gateway))
        return self.post(url, stream=True)

    def uart_gateway(self, gateway, serial_options, test_runner):
        """
            Open a connection to gateway serial port
        """
        url = 'gateway/{}/uart-duck'.format(quote(gateway))

        if test_runner == 'none':
            test_runner = None
        json_data = {
            'serial_options': serial_options,
            'test_runner': test_runner,
        }
        return self.post(url, json=json_data)

    def rename_gateway(self, gateway, new_name):
        """
            Rename a gateway
        """
        url = 'gateway/{}/rename'.format(quote(gateway))
        return self.post(url, json={'name': new_name})

    def start_local_gdb_tunnel(self, gateway, fork):
        """
            Start the local gdb tunnel on gateway
        """
        url = 'gateway/{}/local-gdb'.format(quote(gateway))
        return self.post(url, json={'fork': fork})

    def gpio_set(self, gateway, gpio, type_, pull):
        """
            Set a GPIO pin to input or output
        """
        url = 'gateway/{}/gpio/set'.format(quote(gateway))
        return self.post(url, json={'gpio': gpio, 'type': type_, 'pull': pull})

    def gpio_input(self, gateway, gpio):
        """
            Read from the GPIO pin
        """
        url = 'gateway/{}/gpio/input'.format(quote(gateway))
        return self.post(url, json={'gpio': gpio})

    def gpio_output(self, gateway, gpio, level):
        """
            Write to the GPIO pin
        """
        url = 'gateway/{}/gpio/output'.format(quote(gateway))
        return self.post(url, json={'gpio': gpio, 'level': level})

    def gpio_servo(self, gateway, gpio, pulsewidth, stop):
        """
            Control a servo with GPIO
        """
        url = 'gateway/{}/gpio/servo'.format(quote(gateway))
        return self.post(url, json={'gpio': gpio, 'pulsewidth': pulsewidth, 'stop': stop})

    def gpio_trigger(self, gateway, gpio, pulse_length, level):
        """
            Send a trigger pulse on GPIO
        """
        url = 'gateway/{}/gpio/trigger'.format(quote(gateway))
        return self.post(url, json={'gpio': gpio, 'pulse_length': pulse_length, 'level': level})

    def gpio_hardware_pwm(self, gateway, frequency, dutycycle):
        """
            Start hardware PWM on gpio
        """
        url = 'gateway/{}/gpio/hardware-pwm'.format(quote(gateway))
        return self.post(url, json={'frequency': frequency, 'dutycycle': dutycycle})

    def gpio_hardware_clock(self, gateway, frequency):
        """
            Start hardware clock on gpio
        """
        url = 'gateway/{}/gpio/hardware-clock'.format(quote(gateway))
        return self.post(url, json={'frequency': frequency})

class LagerContext:  # pylint: disable=too-few-public-methods
    """
        Lager Context manager
    """
    def __init__(self, ctx, auth, defaults, debug, style):
        host = os.getenv('LAGER_HOST', _DEFAULT_HOST)
        ws_host = os.getenv('LAGER_WS_HOST', _DEFAULT_WEBSOCKET_HOST)
        base_url = '{}{}'.format(host, '/api/v1/')

        self.session = LagerSession(ctx, auth, base_url=base_url)
        self.defaults = defaults
        self.style = style
        self.ws_host = ws_host
        self.debug = debug
        if auth:
            self.auth_token = auth['token']

    @property
    def default_gateway(self):
        """
            Get default gateway id from config
        """
        return self.defaults.get('gateway_id')

    @default_gateway.setter
    def default_gateway(self, gateway_id):
        self.defaults['gateway_id'] = str(gateway_id)

    def websocket_connection_params(self, socktype, **kwargs):
        """
            Yields a websocket connection to the given path
        """
        if socktype == 'job':
            path = f'/ws/job/{kwargs["job_id"]}'
        elif socktype == 'gdb-tunnel':
            path = f'/ws/gateway/{kwargs["gateway_id"]}/gdb-tunnel'
        else:
            raise ValueError(f'Invalid websocket type: {socktype}')
        uri = urllib.parse.urljoin(self.ws_host, path)

        headers = [
            (b'authorization', self.session.headers['Authorization'].encode()),
        ]
        ctx = get_ssl_context()

        return (uri, dict(extra_headers=headers, ssl_context=ctx))

def get_default_gateway(ctx):
    """
        Check for a default gateway in config; if not present, check if the user
        only has 1 gateway. If so, use that one.
    """
    name = os.getenv('LAGER_GATEWAY')
    if name is None:
        name = ctx.obj.default_gateway

    if name is None:
        session = ctx.obj.session
        resp = session.list_gateways()
        resp.raise_for_status()
        gateways = resp.json()['gateways']

        if not gateways:
            click.secho('No gateways found! Please contact support@lagerdata.com', fg='red')
            ctx.exit(1)
        if len(gateways) == 1:
            ctx.obj.default_gateway = gateways[0]['id']
            return gateways[0]['id']

        hint = 'NAME. Set a default using `lager set default gateway <id>`'
        raise click.MissingParameter(
            param=ctx.command.params[0],
            param_hint=hint,
            ctx=ctx,
            param_type='argument',
        )
    return name

def get_ssl_context():
    """
        Get an SSL context, with custom CA cert if necessary
    """
    cafile_path = os.getenv('LAGER_CAFILE_PATH')
    if not cafile_path:
        # Use default system CA certs
        return None
    ctx = ssl.create_default_context()
    ctx.load_verify_locations(cafile=cafile_path)
    return ctx

def ensure_debugger_running(gateway, ctx):
    """
        Ensure debugger is running on a given gateway
    """
    session = ctx.obj.session
    gateway_status = session.gateway_status(gateway).json()
    if not gateway_status['running']:
        click.secho('Gateway debugger is not running. Please use `lager connect` to run it', fg='red', err=True)
        ctx.exit(1)

class CIEnvironment(Enum):
    """
        Enum representing supported CI system
    """
    HOST = 'host'
    CONTAINER_CI = 'container_ci'

def get_ci_environment():
    """
        Determine whether we are running in CI or not
    """

    if os.getenv('CI') == 'true':
        if os.getenv('DRONE') == 'true' or os.getenv('GITHUB_RUN_ID') or os.getenv('BITBUCKET_BUILD_NUMBER'):
            return CIEnvironment.CONTAINER_CI

    return CIEnvironment.HOST
